from wcd_device_recognizer.services import registry, request_resolver
from wcd_jwt.contrib.registry.services.pipeline import flat_pairs

from . import connector


def connect_interlocutor(context):
    request, pairs = context.get('request'), context.get('registered_pairs')

    if request is None:
        return

    tokens = flat_pairs(pairs)
    interlocutor_dto = request_resolver.resolve(request)
    (interlocutor, network), = registry.register_interlocutors((
        interlocutor_dto,
    ))
    connections = connector.connect(
        interlocutor, tokens=tokens, network=network
    )

    return {
        'interlocutor': interlocutor,
        'interlocutor_dto': interlocutor_dto,
        'interlocutor_network': network,
        'interlocutor_connections': connections,
    }



from django.utils.translation import pgettext_lazy

from rest_framework import serializers
from rest_framework.exceptions import ValidationError
from rest_framework_simplejwt.tokens import Token as TokenStruct

from wcd_jwt.serializers import TokenRefreshSerializer as _TokenRefreshSerializer
from wcd_jwt.services import token_resolver

from .services import token_manager


__all__ = 'TokenExpirationCheckMixin', 'TokenRefreshSerializer',


class TokenExpirationCheckMixin:
    default_error_messages = {
        'token_expired': pgettext_lazy(
            'wcd_jwt', 'Token is not longer valid.',
        )
    }

    def is_expired(self, token: TokenStruct):
        return token_manager.check_expired(token)

    def validate(self, attrs):
        data = super().validate(attrs)

        # TODO: Rework this as there could be no field with that name:
        token = data.get('refresh')

        if token is not None:
            if self.is_expired(token):
                raise ValidationError(
                    self.error_messages['token_expired'], 'token_expired',
                )

        return data


class TokenRefreshSerializer(TokenExpirationCheckMixin, _TokenRefreshSerializer):
    pass


class TokenExpireSerializer(serializers.Serializer):
    default_error_messages = {
        'token_expired': pgettext_lazy(
            'wcd_jwt', 'Token is not longer valid.',
        )
    }

    token = serializers.CharField(required=True)
    affect_tree = serializers.BooleanField(required=False, default=False)

    def validate_token(self, value):
        token = token_resolver.resolve_valid(value, silent=True)

        if token is None or token_manager.check_expired(token, silent=True):
            raise ValidationError(
                self.error_messages['token_expired'], 'token_expired',
            )

        return token

    def commit(self):
        token_manager.expire(
            (self.validated_data['token'],),
            affect_tree=self.validated_data['affect_tree'],
        )

        return {}



from datetime import datetime
from typing import List, Optional, Sequence, Union
from uuid import uuid4, UUID
from django.db import models
from django.utils.translation import pgettext_lazy
from rest_framework_simplejwt.utils import aware_utcnow
from django.conf import settings


__all__ = 'Token', 'TokenUserConnection',


TokenID = Union[str, UUID]


class UUIDable(models.Model):
    class Meta:
        abstract = True

    id = models.UUIDField(
        verbose_name=pgettext_lazy('wcd_jwt', 'ID'),
        primary_key=True, default=uuid4,
    )

    created_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_jwt', 'Created at'),
        auto_now_add=True
    )
    updated_at = models.DateTimeField(
        verbose_name=pgettext_lazy('wcd_jwt', 'Updated at'),
        auto_now=True
    )


class TokenQuerySet(models.QuerySet):
    def relatives(self, ids: Sequence[TokenID]):
        return self.filter(
            models.Q(parent_id__in=ids) | models.Q(id__in=ids)
        )

    def expired(self, now: Optional[datetime] = None):
        now = now if now else aware_utcnow()

        return self.filter(
            models.Q(expired_at__lt=now)
            |
            models.Q(internal_expired_at__lt=now)
        )

    def active(self, now: Optional[datetime] = None):
        now = now if now else aware_utcnow()

        return self.filter(
            models.Q(expired_at__gte=now)
            &
            models.Q(internal_expired_at__gte=now)
        )

    def set_expired(self, now: Optional[datetime] = None):
        return self.update(expired_at=now if now else aware_utcnow())

    def collect_tree(self, ids: Sequence[TokenID]) -> List[TokenID]:
        count_first = 0
        ids = set(ids) - {None}
        count_second = len(ids)

        while count_first != count_second:
            count_first = count_second
            relatives = (
                x
                for items in self.relatives(ids).values_list('id', 'parent_id')
                for x in items
            )
            ids = set(relatives) - {None}
            count_second = len(ids)

        return list(ids)


class Token(UUIDable):
    # type: models.Manager[TokenQuerySet]
    objects = TokenQuerySet.as_manager()

    class Meta:
        verbose_name = pgettext_lazy('wcd_jwt', 'Token')
        verbose_name_plural = pgettext_lazy('wcd_jwt', 'Tokens')
        ordering = ('-created_at',)

    jti = models.CharField(
        verbose_name=pgettext_lazy('wcd_jwt', 'JTI'), max_length=512,
        db_index=True,
    )
    token = models.TextField(
        verbose_name=pgettext_lazy('wcd_jwt', 'Token'), db_index=True,
    )

    parent = models.ForeignKey(
        'self', related_name='children', on_delete=models.SET_NULL,
        verbose_name=pgettext_lazy('wcd_jwt', 'Parent token'),
        null=True, blank=True,
    )

    expired_at = models.DateTimeField(pgettext_lazy('wcd_jwt', 'Expired at'))
    internal_expired_at = models.DateTimeField(
        pgettext_lazy('wcd_jwt', 'Expired at(internal)'), editable=False,
    )

    def __str__(self):
        return f'{self.pk}'


class TokenUserConnection(UUIDable):
    class Meta:
        verbose_name = pgettext_lazy('wcd_jwt', 'User token connection')
        verbose_name_plural = pgettext_lazy('wcd_jwt', 'User token connections')
        unique_together = (
            ('user', 'token',),
        )

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name='token_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_jwt', 'Connected user'),
    )
    token = models.ForeignKey(
        Token,
        related_name='user_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_jwt', 'Connected token'),
    )

    def __str__(self):
        return f'User #{self.user_id} token #{self.token_id}'



from typing import List, Sequence, Optional
from wcd_device_recognizer.models import Interlocutor, InterlocutorNetwork
from wcd_device_recognizer.utils import model_bulk_get_or_create
from wcd_jwt.contrib.registry.models import Token

from ..models import TokenInterlocutorConnection
from ..signals import interlocutor_connected


def connect(
    interlocutor: Interlocutor,
    tokens: Sequence[Token],
    network: Optional[InterlocutorNetwork] = None,
) -> List[TokenInterlocutorConnection]:
    connections = model_bulk_get_or_create(TokenInterlocutorConnection, [
        ({'token': token}, {'interlocutor': interlocutor, 'network': network})
        for token in tokens
    ])

    interlocutor_connected.send(
        None, connections=connections, interlocutor=interlocutor,
        network=network, tokens=tokens,
    )

    return connections



import pytest
import time
from django.test import RequestFactory
from django.views.decorators.csrf import csrf_exempt

from wcd_jwt.views import token_obtain_view, token_refresh_view
from wcd_jwt.contrib.registry.models import Token


obtain_view = csrf_exempt(token_obtain_view)
refresh_view = csrf_exempt(token_refresh_view)


@pytest.mark.django_db
def test_obtain_tokens_registered(make_user, rf: RequestFactory):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})
    response = obtain_view(r)

    assert 2 == Token.objects.filter(token__in=[
        response.data['refresh'], response.data['access']
    ]).count()


@pytest.mark.django_db
def test_refresh_tokens_registered(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    assert 'refresh' not in refreshed
    assert 1 == Token.objects.filter(token__in=[
        refreshed['access']
    ]).count()
    assert 3 == Token.objects.count()

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refreshed2 = refresh_view(
            rf.post('/', {'refresh': obtained['refresh']})
        ).data

        assert 2 == Token.objects.filter(token__in=[
            refreshed2['refresh'], refreshed2['access']
        ]).count()
        assert 5 == Token.objects.count()

    parents = tuple(sorted([
        (obtained['refresh'], None),
        (obtained['access'], obtained['refresh']),
        (refreshed['access'], obtained['refresh']),
        (refreshed2['access'], refreshed2['refresh']),
        (refreshed2['refresh'], obtained['refresh']),
    ]))

    result = tuple(sorted(Token.objects.values_list('token', 'parent__token')))

    assert parents == result


@pytest.mark.django_db
def test_register_parallel(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})
    run = {'yes': False}

    class RT:
        def __init__(self, target, args, kwargs, **kw):
            self.target = target
            self.args = args
            self.kwargs = kwargs

        def start(self):
            run['yes'] = True
            self.target(*self.args, **self.kwargs)

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.contrib.registry.utils.threading.Thread', RT)
        response = obtain_view(r)
        assert not run['yes']

        m.setattr('wcd_jwt.contrib.registry.conf.settings.TOKEN_REGISTRATION_ON_SIGNAL_PARALLEL', True)
        response = obtain_view(r)
        assert run['yes']

    assert 4 == Token.objects.count()


@pytest.mark.django_db
def test_disabled_autoregister(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.contrib.registry.conf.settings.TOKEN_REGISTRATION_ON_SIGNAL', False)
        response = obtain_view(r)
        refreshed = refresh_view(
            rf.post('/', {'refresh': response.data['refresh']})
        ).data

    assert 0 == Token.objects.count()



from typing import Tuple
from px_pipeline import Filter, StraightPipeline
from functools import lru_cache
from django.contrib.auth import get_user_model

from ..conf import settings
from ..signals import registration_pipeline_run

from . import token_manager, user_connector


registration_filter = Filter()


def flat_pairs(pairs):
    return {y for x in pairs for y in x}


@lru_cache
def _cached_pipeline(pipeline: Tuple):
    return StraightPipeline(pipeline)


def register_pairs(context):
    pairs = context.get('pairs')

    if pairs is None:
        return

    return {'registered_pairs': token_manager.register(pairs)}


def connecting_user(context):
    user, pairs = context.get('user'), context.get('registered_pairs')

    if len(pairs) == 0:
        return

    flattened_pairs = flat_pairs(pairs)

    if user is None:
        # Searching for user connections of the previous token(s).
        # To connect them. If there is more than one user - something
        # wrong happened and no connection will be established.
        users = list(
            get_user_model().objects
            .filter(token_connections__token__in=flattened_pairs)
        )

        if len(users) != 1:
            return

        user = users[0]

    return {'connections': user_connector.connect(user, flattened_pairs)}


def run(context: dict) -> dict:
    result = registration_filter(
        _cached_pipeline(tuple(settings.TOKEN_REGISTRATION_PIPELINE))(context)
    )

    registration_pipeline_run.send(None, result=result)

    return result



# Generated by Django 2.2 on 2022-07-21 20:45

from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ('wcd_jwt_registry', '0001_initial'),
        ('wcd_device_recognizer', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='TokenInterlocutorConnection',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('interlocutor', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='token_connections', to='wcd_device_recognizer.Interlocutor', verbose_name='Connected interlocutor')),
                ('network', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='token_connections', to='wcd_device_recognizer.InterlocutorNetwork', verbose_name='Connected interlocutor network')),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='interlocutor_connections', to='wcd_jwt_registry.Token', verbose_name='Connected token')),
            ],
            options={
                'verbose_name': 'Interlocutor token connection',
                'verbose_name_plural': 'Interlocutor token connections',
            },
        ),
    ]



from dataclasses import dataclass, field
from typing import Sequence
from px_pipeline.pipelines import Executable
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


@s('WCD_JWT_REGISTRY')
@dataclass
class Settings:
    TOKEN_EXPIRE_SERIALIZER: str = 'wcd_jwt.contrib.registry.serializers.TokenExpireSerializer'
    TOKEN_EXPIRE_ON_REFRESH: bool = False

    TOKEN_REGISTRATION_PIPELINE: Sequence[Executable] = field(default_factory=lambda: [
        'wcd_jwt.contrib.registry.services.pipeline.register_pairs',
        'wcd_jwt.contrib.registry.services.pipeline.connecting_user',
    ])
    TOKEN_REGISTRATION_ON_SIGNAL: bool = True
    TOKEN_REGISTRATION_ON_SIGNAL_PARALLEL: bool = False


settings = Settings()



import pytest
from django.test import RequestFactory
from django.views.decorators.csrf import csrf_exempt

from wcd_jwt.views import token_obtain_view, token_refresh_view
from wcd_jwt.contrib.registry.models import Token, TokenUserConnection


obtain_view = csrf_exempt(token_obtain_view)
refresh_view = csrf_exempt(token_refresh_view)


@pytest.mark.django_db
def test_refresh_tokens_registered(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    assert 1 == TokenUserConnection.objects.filter(token__token__in=[
        refreshed['access']
    ]).count()
    assert 3 == TokenUserConnection.objects.count()
    assert 3 == TokenUserConnection.objects.filter(user=user).count()

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refreshed2 = refresh_view(
            rf.post('/', {'refresh': obtained['refresh']})
        ).data

        assert 2 == TokenUserConnection.objects.filter(token__token__in=[
            refreshed2['refresh'], refreshed2['access']
        ]).count()
        assert 5 == TokenUserConnection.objects.count()
        assert 5 == TokenUserConnection.objects.filter(user=user).count()



from django.contrib import admin
from django.utils.translation import pgettext_lazy

from .models import TokenInterlocutorConnection


@admin.register(TokenInterlocutorConnection)
class TokenInterlocutorConnectionAdmin(admin.ModelAdmin):
    list_display = (
        'id',
        'get_users', 'get_device', 'get_os', 'get_app',
        'interlocutor', 'token', 'created_at',
    )
    date_hierarchy = 'created_at'
    list_select_related = (
        'interlocutor', 'interlocutor__device', 'interlocutor__os',
        'interlocutor__app',
        'token',
    )
    readonly_fields = 'updated_at', 'created_at',
    autocomplete_fields = 'interlocutor', 'token',
    search_fields = (
        'id', 'interlocutor__user_agent', 'token__token', 'token__jti',
        'token__user_connections__user__pk',
        'token__user_connections__user__username',
    )

    def get_queryset(self, request):
        return (
            super().get_queryset(request)
            .prefetch_related(
                'token__user_connections',
                'token__user_connections__user',
            )
        )

    def get_users(self, obj):
        if not obj.token_id:
            return ''

        return ', '.join(str(x.user) for x in obj.token.user_connections.all())
    get_users.short_description = pgettext_lazy('wcd_jwt', 'Users')

    def get_device(self, obj):
        return obj.interlocutor.device
    get_device.short_description = pgettext_lazy('wcd_jwt', 'Device')
    get_device.admin_order_field = 'interlocutor__device'

    def get_os(self, obj):
        return obj.interlocutor.os
    get_os.short_description = pgettext_lazy('wcd_jwt', 'OS')
    get_os.admin_order_field = 'interlocutor__os'

    def get_app(self, obj):
        return obj.interlocutor.app
    get_app.short_description = pgettext_lazy('wcd_jwt', 'App')
    get_app.admin_order_field = 'interlocutor__app'

    def get_queryset(self, request):
        return super().get_queryset(request).distinct()



import pytest
from django.test import RequestFactory

from wcd_jwt.views import token_obtain_view, token_refresh_view
from django.views.decorators.csrf import csrf_exempt


obtain_view = csrf_exempt(token_obtain_view)
refresh_view = csrf_exempt(token_refresh_view)


@pytest.mark.django_db
def test_obtain(make_user, rf: RequestFactory):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})

    response = obtain_view(r)

    assert 'access' in response.data
    assert 'refresh' in response.data


@pytest.mark.django_db
def test_refresh(make_user, rf: RequestFactory, monkeypatch, django_assert_num_queries):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    )
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained.data['refresh']})
    )

    assert 'access' in refreshed.data
    assert 'refresh' not in refreshed.data

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        m.setattr('wcd_jwt.contrib.registry.conf.settings.TOKEN_REGISTRATION_PIPELINE', [
            'wcd_jwt.contrib.registry.services.pipeline.register_pairs',
            'wcd_jwt.contrib.registry.services.pipeline.connecting_user',
        ])

        # TODO: There is too much queries happening here.
        # Should rework it somehow.
        with django_assert_num_queries(9):
            refreshed2 = refresh_view(
                rf.post('/', {'refresh': obtained.data['refresh']})
            )

        assert 'access' in refreshed2.data
        assert 'refresh' in refreshed2.data
        assert refreshed2.data['refresh'] != obtained.data['refresh']

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)

        # TODO: There is too much queries happening here.
        # Should rework it somehow.
        with django_assert_num_queries(16):
            refreshed3 = refresh_view(
                rf.post('/', {'refresh': refreshed2.data['refresh']})
            )



import pytest
from django.test import RequestFactory
from django.views.decorators.csrf import csrf_exempt

from wcd_jwt.views import token_obtain_view, token_refresh_view
from wcd_jwt.contrib.device_registry.models import TokenInterlocutorConnection
from wcd_device_recognizer.models import Interlocutor


obtain_view = csrf_exempt(token_obtain_view)
refresh_view = csrf_exempt(token_refresh_view)


@pytest.mark.django_db
def test_register_interlocutor(make_user, rf: RequestFactory):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})
    response = obtain_view(r)

    assert 2 == TokenInterlocutorConnection.objects.count()
    assert 1 == Interlocutor.objects.count()



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('JWTDeviceRegistryConfig',)


class JWTDeviceRegistryConfig(AppConfig):
    name = 'wcd_jwt.contrib.device_registry'
    label = 'wcd_jwt_device_registry'
    verbose_name = pgettext_lazy('wcd_jwt', 'JWT Token device registry')



from django.apps import AppConfig
from django.utils.translation import pgettext_lazy


__all__ = ('JWTRegistryConfig',)


class JWTRegistryConfig(AppConfig):
    name = 'wcd_jwt.contrib.registry'
    label = 'wcd_jwt_registry'
    verbose_name = pgettext_lazy('wcd_jwt', 'JWT Token registry')

    def ready(self):
        from . import subscriptions



from django.contrib import admin
from django.utils.translation import pgettext_lazy
from rest_framework_simplejwt.tokens import AccessToken as TokenClass

from .models import Token, TokenUserConnection


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):
    list_display = 'id', 'jti', 'get_token_type', 'parent_id', 'expired_at', 'created_at',
    date_hierarchy = 'created_at'
    autocomplete_fields = 'parent',
    search_fields = 'id', 'jti', 'token', 'user_connections__user__username'
    readonly_fields = (
        'get_token_type', 'get_token_payload',
        'internal_expired_at', 'updated_at', 'created_at',
    )
    fields = (
        'id', 'jti',
        'parent',
        'token',
        'get_token_type',
        'get_token_payload',
        ('expired_at', 'internal_expired_at',),
        ('updated_at', 'created_at',),
    )

    def get_token_payload(self, obj):
        token = TokenClass(obj.token, verify=False)

        return token.payload
    get_token_payload.short_description = pgettext_lazy('wcd_jwt', 'Payload')

    def get_token_type(self, obj):
        payload = self.get_token_payload(obj)

        return payload.get('token_type', '')
    get_token_type.short_description = pgettext_lazy('wcd_jwt', 'Type')


@admin.register(TokenUserConnection)
class TokenUserConnectionAdmin(admin.ModelAdmin):
    list_display = 'id', 'user', 'token', 'created_at',
    date_hierarchy = 'created_at'
    list_select_related = 'user', 'token',
    autocomplete_fields = 'user', 'token',
    search_fields = 'id', 'user__username', 'token__token', 'token__jti',
    readonly_fields = 'updated_at', 'created_at',



from django.db import models
from django.utils.translation import pgettext_lazy

from wcd_jwt.contrib.registry.models import UUIDable, Token
from wcd_device_recognizer.models import Interlocutor, InterlocutorNetwork


__all__ = 'TokenInterlocutorConnection',


class TokenInterlocutorConnection(UUIDable):
    class Meta:
        verbose_name = pgettext_lazy('wcd_jwt', 'Interlocutor token connection')
        verbose_name_plural = pgettext_lazy('wcd_jwt', 'Interlocutor token connections')
        ordering = ('-created_at',)

    interlocutor = models.ForeignKey(
        Interlocutor,
        related_name='token_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_jwt', 'Connected interlocutor'),
    )
    network = models.ForeignKey(
        InterlocutorNetwork,
        related_name='token_connections', on_delete=models.SET_NULL,
        verbose_name=pgettext_lazy('wcd_jwt', 'Connected interlocutor network'),
        null=True, blank=True,
    )
    token = models.ForeignKey(
        Token,
        related_name='interlocutor_connections', on_delete=models.CASCADE,
        verbose_name=pgettext_lazy('wcd_jwt', 'Connected token'),
    )

    def __str__(self):
        return f'Interlocutor #{self.interlocutor_id} token #{self.token_id}'



from django.dispatch import receiver
from django.contrib.auth.models import update_last_login

from .signals import token_obtained, token_refreshed
from .conf import settings


@receiver(token_obtained)
def update_last_login_on_token_obtain(sender, serializer, **kwargs):
    if settings.UPDATE_LAST_LOGIN:
        update_last_login(None, serializer.validated_data['user'])



from django.http import HttpRequest
from django.contrib.auth.models import AnonymousUser
from rest_framework.authentication import BaseAuthentication
from rest_framework_simplejwt.exceptions import (
    AuthenticationFailed, InvalidToken,
)

from .utils import cached_import_string
from .conf import settings


def resolve_user(request: HttpRequest):
    AuthenticationClass: BaseAuthentication = cached_import_string(
        settings.AUTHENTICATION_CLASS
    )
    result = None

    try:
        result = AuthenticationClass().authenticate(request)
    except (InvalidToken, AuthenticationFailed) as e:
        pass

    if result is None:
        return AnonymousUser(), None

    return result



from datetime import datetime
from typing import Optional

from rest_framework_simplejwt.utils import aware_utcnow

from ..models import Token


def archive(now: Optional[datetime] = None):
    now = aware_utcnow() if now is None else now

    return Token.objects.filter(internal_expired_at__lte=now).delete()



# Generated by Django 2.2 on 2022-07-22 07:27

import datetime
from django.db import migrations, models
from django.utils.timezone import utc


class Migration(migrations.Migration):

    dependencies = [
        ('wcd_jwt_registry', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='token',
            name='internal_expired_at',
            field=models.DateTimeField(default=datetime.datetime(2022, 7, 23, 7, 27, 51, 800647, tzinfo=utc), editable=False, verbose_name='Expired at'),
            preserve_default=False,
        ),
    ]



from django.contrib import admin
from django.contrib.auth import get_user_model
from django.conf import settings

from jet.filters import RelatedFieldAjaxListFilter


User = get_user_model()

if 'wcd_jwt.contrib.registry' in settings.INSTALLED_APPS:
    from wcd_jwt.contrib.registry.models import Token, TokenUserConnection
    from wcd_jwt.contrib.registry.admin import TokenAdmin, TokenUserConnectionAdmin


    admin.site.unregister(Token)
    admin.site.unregister(TokenUserConnection)


    @admin.register(Token)
    class TokenWithFiltersAdmin(TokenAdmin):
        list_filter = (
            ('user_connections__user', RelatedFieldAjaxListFilter),
        )


    @admin.register(TokenUserConnection)
    class TokenUserConnectionWithFiltersAdmin(TokenUserConnectionAdmin):
        list_filter = (
            ('user', RelatedFieldAjaxListFilter),
            ('token', RelatedFieldAjaxListFilter),
        )


    if not hasattr(User, 'autocomplete_search_fields'):
        User.autocomplete_search_fields = staticmethod(lambda *a: ('username',))

    if not hasattr(Token, 'autocomplete_search_fields'):
        Token.autocomplete_search_fields = staticmethod(lambda *a: ('pk', 'token', 'jti',))


if 'wcd_jwt.contrib.device_registry' in settings.INSTALLED_APPS:
    from wcd_device_recognizer.models import Interlocutor, InterlocutorNetwork

    if not hasattr(Interlocutor, 'autocomplete_search_fields'):
        Interlocutor.autocomplete_search_fields = staticmethod(lambda *a: ('outer_id', 'user_agent',))
    if not hasattr(InterlocutorNetwork, 'autocomplete_search_fields'):
        InterlocutorNetwork.autocomplete_search_fields = staticmethod(lambda *a: ('interlocutor__user_agent', 'ip',))



from django.http import HttpRequest
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import SimpleLazyObject

from .resolver import resolve_user


class AuthenticationMiddleware(MiddlewareMixin):
    def process_request(self, request: HttpRequest):
        # Default django's authentication middleware could already
        # authenticate user. So there is no sense to do it by our own.
        if hasattr(request, 'user'):
            if request.user.is_authenticated:
                return

        request.user = SimpleLazyObject(lambda: resolve_user(request)[0])



from dataclasses import dataclass, field
from functools import partial
from typing import Optional, Sequence
from px_settings.contrib.django import settings as s


__all__ = 'Settings', 'settings',


def get_simplejwt_setting(key: str, default: Optional[str] = None):
    result = default

    if hasattr(simple_jwt_settings, key):
        result = getattr(simple_jwt_settings, key)

    from rest_framework_simplejwt.settings import api_settings

    return (
        result
        if result is not None else
        default
        if default is not None else
        getattr(api_settings, key)
    )


pjwt = partial(partial, get_simplejwt_setting)


@s('SIMPLE_JWT')
@dataclass
class SimpleJWTSettings:
    TOKEN_OBTAIN_SERIALIZER: Optional[str] = None
    TOKEN_REFRESH_SERIALIZER: Optional[str] = None
    ROTATE_REFRESH_TOKENS: Optional[bool] = None
    UPDATE_LAST_LOGIN: Optional[bool] = None


@s('WCD_JWT')
@dataclass
class Settings:
    TOKEN_OBTAIN_SERIALIZER: str = field(default_factory=pjwt(
        'TOKEN_OBTAIN_SERIALIZER',
        'wcd_jwt.serializers.TokenObtainPairSerializer',
    ))
    TOKEN_REFRESH_SERIALIZER: str = field(default_factory=pjwt(
        'TOKEN_REFRESH_SERIALIZER',
        'wcd_jwt.serializers.TokenRefreshSerializer',
    ))
    ROTATE_REFRESH_TOKENS: bool = field(
        default_factory=pjwt('ROTATE_REFRESH_TOKENS')
    )
    UPDATE_LAST_LOGIN: bool = field(
        default_factory=pjwt('UPDATE_LAST_LOGIN')
    )
    TOKEN_TYPES: Sequence[str] = field(default_factory=lambda: [
        'rest_framework_simplejwt.tokens.AccessToken',
        'rest_framework_simplejwt.tokens.RefreshToken',
    ])

    AUTHENTICATION_CLASS: str = 'wcd_jwt.authentication.JWTAuthentication'


simple_jwt_settings = SimpleJWTSettings()
settings = Settings()



from django.dispatch import receiver
from wcd_jwt.signals import token_refreshed, token_obtained

from .services import pipeline, token_manager
from .utils import may_parallel
from .conf import settings


@receiver(token_obtained)
@may_parallel
def register_token_on_obtain(sender, serializer, result, request, **kwargs):
    if not settings.TOKEN_REGISTRATION_ON_SIGNAL:
        return

    v = serializer.validated_data

    pipeline.run({
        'pairs': [(v['access'], v['refresh'])],
        'user': v.get('user') or None,
        'request': request,
    })


@receiver(token_refreshed)
@may_parallel
def register_token_on_refresh(sender, serializer, result, request, **kwargs):
    if not settings.TOKEN_REGISTRATION_ON_SIGNAL:
        return

    v = serializer.validated_data
    base = v['refresh']
    pairs = [(v['access'], base)]
    expiration = ((base,), (base,))

    if 'refresh' in result:
        refresh = serializer.token_class(result['refresh'])
        pairs = (
            # Wer'e replacing access token parent with newly generated refresh.
            # It's because we want to form a straight relation between
            # responded refresh->access tokens.
            (v['access'], refresh),
            (refresh, base),
        )
        expiration = ((base,), ())

    if settings.TOKEN_EXPIRE_ON_REFRESH:
        expire, exclude = expiration
        token_manager.expire(expire, affect_tree=True, exclude=exclude)

    pipeline.run({'pairs': pairs, 'request': request})



# Generated by Django 2.2 on 2022-07-19 15:19

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Token',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('jti', models.CharField(db_index=True, max_length=512, verbose_name='JTI')),
                ('token', models.TextField(db_index=True, verbose_name='Token')),
                ('expired_at', models.DateTimeField(verbose_name='Expired at')),
                ('parent', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='children', to='wcd_jwt_registry.Token', verbose_name='Parent token')),
            ],
            options={
                'verbose_name': 'Token',
                'verbose_name_plural': 'Tokens',
            },
        ),
        migrations.CreateModel(
            name='TokenUserConnection',
            fields=[
                ('id', models.UUIDField(default=uuid.uuid4, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, verbose_name='Created at')),
                ('updated_at', models.DateTimeField(auto_now=True, verbose_name='Updated at')),
                ('token', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='user_connections', to='wcd_jwt_registry.Token', verbose_name='Connected token')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='token_connections', to=settings.AUTH_USER_MODEL, verbose_name='Connected user')),
            ],
            options={
                'verbose_name': 'User token connection',
                'verbose_name_plural': 'User token connections',
                'unique_together': {('user', 'token')},
            },
        ),
    ]



from typing import Optional
from django.urls import path
from django.dispatch import Signal
from django.contrib.auth import login
from rest_framework_simplejwt.views import (
    TokenViewBase as _TokenViewBase, TokenVerifyView,
    token_verify as token_verify_view,
)
from django.views.decorators.csrf import csrf_exempt
from rest_framework_simplejwt.exceptions import InvalidToken, TokenError
from rest_framework.response import Response
from rest_framework import status

from .conf import settings
from .signals import token_obtained, token_refreshed


__all__ = (
    'TokenRefreshView', 'TokenObtainView', 'TokenVerifyView',
    'token_refresh_view', 'token_obtain_view', 'token_verify_view',
    'make_urlpatterns',
)


class TokenViewBase(_TokenViewBase):
    notify_signal: Optional[Signal] = None

    def commit(self, serializer) -> dict:
        return (
            serializer.commit()
            if hasattr(serializer, 'commit')
            else serializer.validated_data
        )

    def notify(self, result: dict, serializer, request):
        if self.notify_signal is not None:
            self.notify_signal.send(
                self.__class__, result=result, serializer=serializer,
                request=request,
            )

    def post(self, request, *args, **kwargs):
        self.serializer = self.get_serializer(data=request.data)

        try:
            self.serializer.is_valid(raise_exception=True)
        except TokenError as e:
            raise InvalidToken(e.args[0])

        result = self.commit(self.serializer)
        self.notify(result=result, serializer=self.serializer, request=request)

        return Response(result, status=status.HTTP_200_OK)


class TokenRefreshView(TokenViewBase):
    """
    Takes a refresh type JSON web token and returns an access type JSON web
    token if the refresh token is valid.
    """
    notify_signal = token_refreshed
    _serializer_class = settings.TOKEN_REFRESH_SERIALIZER


class TokenObtainView(TokenViewBase):
    """
    Takes a set of user credentials and returns a JSON web token to prove the
    authentication of those credentials.
    """
    notify_signal = token_obtained
    _serializer_class = settings.TOKEN_OBTAIN_SERIALIZER


class WithSessionTokenObtainView(TokenObtainView):
    """
    Takes a set of user credentials and returns a JSON web token to prove the
    authentication of those credentials.
    """
    def commit(self, serializer) -> dict:
        if self.serializer.is_valid():
            login(self.request, serializer.user)

        return super().commit(serializer)


token_refresh_view = TokenRefreshView.as_view()
token_obtain_view = TokenObtainView.as_view()
with_session_token_obtain_view = WithSessionTokenObtainView.as_view()


def make_urlpatterns(
    obtain_view=with_session_token_obtain_view,
    refresh_view=token_refresh_view,
    verify_view=token_verify_view,
):
    return [
        path('obtain/', csrf_exempt(obtain_view), name='obtain'),
        path('refresh/', csrf_exempt(refresh_view), name='refresh'),
        path('verify/', csrf_exempt(verify_view), name='verify'),
    ]



from functools import lru_cache
from django.utils.module_loading import import_string
from rest_framework_simplejwt.tokens import Token as TokenStruct
from rest_framework_simplejwt.settings import api_settings


@lru_cache
def cached_import_string(path: str):
    return import_string(path)


def get_token_jti(token: TokenStruct) -> str:
    return token.payload[api_settings.JTI_CLAIM]



import threading
from functools import wraps

from .conf import settings


def maybe_parallel_execution(callback, args, kwargs):
    if settings.TOKEN_REGISTRATION_ON_SIGNAL_PARALLEL:
        thread = threading.Thread(target=callback, args=args, kwargs=kwargs, daemon=True)
        thread.start()
    else:
        callback(*args, **kwargs)


def may_parallel(callback):
    @wraps(callback)
    def wrapper(*a, **kw):
        return maybe_parallel_execution(callback, a, kw)

    return wrapper



import pytest
from datetime import timedelta
from django.test import RequestFactory
from django.views.decorators.csrf import csrf_exempt
from django.utils import timezone
from rest_framework_simplejwt.tokens import AccessToken
from rest_framework_simplejwt.utils import aware_utcnow

from wcd_jwt.views import token_obtain_view, token_refresh_view
from wcd_jwt.contrib.registry.models import Token
from wcd_jwt.contrib.registry.views import token_expire_view
from wcd_jwt.contrib.registry.services import token_manager, archiver


obtain_view = csrf_exempt(token_obtain_view)
refresh_view = csrf_exempt(token_refresh_view)
expire_view = csrf_exempt(token_expire_view)


@pytest.mark.django_db
def test_expire_part(make_user, rf: RequestFactory):
    user, pswd = make_user('user')
    r = rf.post('/', {'username': user.username, 'password': pswd})
    response = obtain_view(r).data

    assert 0 == Token.objects.expired().count()

    response2 = obtain_view(r).data

    token_manager.expire([AccessToken(response2['access'])], affect_tree=True)

    assert 2 == Token.objects.expired().count()
    assert 0 == (
        Token.objects
        .filter(token__in=[response['access'], response['refresh']])
        .expired().count()
    )


@pytest.mark.django_db
def test_expire_large_tree(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refresh_view(rf.post('/', {'refresh': obtained['refresh']}))

    assert 0 == Token.objects.expired().count()
    assert 5 == Token.objects.count()
    token_manager.expire([AccessToken(obtained['access'])])
    assert 1 == Token.objects.expired().count()
    token_manager.expire([AccessToken(refreshed['access'])], affect_tree=True)
    assert 5 == Token.objects.expired().count()


@pytest.mark.django_db
def test_expire_view(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refresh_view(rf.post('/', {'refresh': obtained['refresh']}))

    assert 0 == Token.objects.expired().count()
    assert 5 == Token.objects.count()
    expire_view(
        rf.post('/', {'token': obtained['access']})
    ).data
    assert 1 == Token.objects.expired().count()
    expire_view(
        rf.post('/', {'token': refreshed['access'], 'affect_tree': True})
    ).data
    assert 5 == Token.objects.expired().count()

    response = expire_view(
        rf.post('/', {'token': obtained['access']})
    )
    assert response.status_code == 400
    assert response.data['token'][0].code == 'token_expired'

    response = expire_view(
        rf.post('/', {'token': 'wrongkoken'})
    )
    assert response.status_code == 400


@pytest.mark.django_db
def test_expire_on_refresh(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.contrib.registry.conf.settings.TOKEN_EXPIRE_ON_REFRESH', True)
        refresh_view(rf.post('/', {'refresh': obtained['refresh']}))

        assert 2 == Token.objects.expired().count()
        assert 2 == Token.objects.active().count()

        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refresh_view(rf.post('/', {'refresh': obtained['refresh']}))

        assert 4 == Token.objects.expired().count()
        assert 2 == Token.objects.active().count()


@pytest.mark.django_db
def test_archiver(make_user, rf: RequestFactory, monkeypatch):
    user, pswd = make_user('user')
    obtained = obtain_view(
        rf.post('/', {'username': user.username, 'password': pswd})
    ).data
    refreshed = refresh_view(
        rf.post('/', {'refresh': obtained['refresh']})
    ).data

    with monkeypatch.context() as m:
        m.setattr('wcd_jwt.conf.settings.ROTATE_REFRESH_TOKENS', True)
        refresh_view(rf.post('/', {'refresh': obtained['refresh']}))

    assert 0 == Token.objects.expired().count()
    assert 5 == Token.objects.count()

    expired = (
        Token.objects
        .filter(token__in=[obtained['refresh'], obtained['access']])
        .update(internal_expired_at=aware_utcnow() - timedelta(seconds=1))
    )

    assert 2 == Token.objects.expired().count()
    assert 5 == Token.objects.count()

    archiver.archive()

    assert 0 == Token.objects.expired().count()
    assert 3 == Token.objects.count()
